/**
 * This file is part of the SMT-LIB 2 preprocessor.
 * <http://www.cprover.org/SMT-LIB-LSM/>
 *
 * Copyright (C) 2009 Philipp Ruemmer <ph_r@gmx.net>
 *
 * The SMT-LIB 2 preprocessor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The SMT-LIB 2 preprocessor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the SMT-LIB 2 preprocessor.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <iostream>

#include "multi-arity-ops.H"
#include "util.H"

namespace Resolver {

using namespace std;

MultiArityFunctionFamily::MultiArityFunctionFamily(const FunctionFamilyPtr& basisFunction)
    : basisFunction(basisFunction) {}

MultiArityFunctionFamily::MultiArityFunctionFamily(const MonoFunctionPtr& basisFunction)
    : basisFunction(new MonoFunctionFamily(basisFunction)) {}

MonoFunctionPtr
    MultiArityFunctionFamily::lookup(const std::vector<std::string>& indexes,
                                     const std::vector<SortPtr>& argumentSorts) {
    if (!indexes.empty())
        // this is currently not supported and not necessary for any of the
        // present theories
        throw NoSuchSymbolException("");

    if (argumentSorts.empty())
        throw NoSuchSymbolException("");

    // check whether the requested instance already exists
    for (vector<MonoFunctionPtr>::iterator it = instances.begin();
         it != instances.end();
         ++it) {
        if (equalVectors((*it)->argumentSorts, argumentSorts))
            return *it;
    }

    const MonoFunctionPtr res = createInstanceFor(argumentSorts);
    instances.push_back(res);

    return res;
}

MonoFunctionPtr
    MultiArityFunctionFamily::lookup(const std::vector<std::string>& indexes,
                                     const SortPtr& resSort,
                                     const std::vector<SortPtr>& argumentSorts) {
    const MonoFunctionPtr res = lookup(indexes, argumentSorts);

    if (!(*res->resultSort == *resSort))
        throw NoSuchSymbolException("");

    return res;
}

////////////////////////////////////////////////////////////////////////////////

class LeftAssocMonoFunction : public ExpandableMonoFunction {
public:
    template <class I> LeftAssocMonoFunction(const IndexedString& name,
                                             SortPtr res, I argBegin, I argEnd,
                                             const FunctionFamilyPtr& basisFunction)
        : ExpandableMonoFunction(name, res, argBegin, argEnd, basisFunction) {}

    ExpressionPtr expand(const std::vector<ExpressionPtr>& arguments) {
        assert(arguments.size() == argumentSorts.size());

        ExpressionPtr res = arguments.front();
        for (int i = 1; i < argumentSorts.size(); ++i)
            res = basisFunction->m(res, arguments[i]);

        return res;
    }
};

////////////////////////////////////////////////////////////////////////////////

class RightAssocMonoFunction : public ExpandableMonoFunction {
public:
    template <class I> RightAssocMonoFunction(const IndexedString& name,
                                              SortPtr res, I argBegin, I argEnd,
                                              const FunctionFamilyPtr& basisFunction)
        : ExpandableMonoFunction(name, res, argBegin, argEnd, basisFunction) {}

    ExpressionPtr expand(const std::vector<ExpressionPtr>& arguments) {
        assert(arguments.size() == argumentSorts.size());

        ExpressionPtr res = arguments.back();
        for (int i = argumentSorts.size() - 2; i >= 0; --i)
            res = basisFunction->m(arguments[i], res);

        return res;
    }
};

////////////////////////////////////////////////////////////////////////////////

class ChainableMonoFunction : public ExpandableMonoFunction {
private:
    const MonoFunctionPtr fun_true;
    const MonoFunctionPtr fun_and;

public:
    template <class I> ChainableMonoFunction(const IndexedString& name,
                                             SortPtr res, I argBegin, I argEnd,
                                             const FunctionFamilyPtr& basisFunction,
                                             const MonoFunctionPtr& fun_true,
                                             const MonoFunctionPtr& fun_and)
        : ExpandableMonoFunction(name, res, argBegin, argEnd, basisFunction),
          fun_true(fun_true), fun_and(fun_and) {}

    ExpressionPtr expand(const std::vector<ExpressionPtr>& arguments) {
        assert(arguments.size() == argumentSorts.size());

        if (arguments.size() == 1)
            return fun_true->m();

        ExpressionPtr res;
        for (int i = 0; i < arguments.size() - 1; ++i) {
            const ExpressionPtr newConjunct =
                basisFunction->m(arguments[i], arguments[i+1]);
            if (res)
                res = fun_and->m(res, newConjunct);
            else
                res = newConjunct;
        }

        return res;
    }
};

////////////////////////////////////////////////////////////////////////////////

LeftAssocFunctionFamily::LeftAssocFunctionFamily(const FunctionFamilyPtr& basisFunction)
    : MultiArityFunctionFamily(basisFunction) {}

LeftAssocFunctionFamily::LeftAssocFunctionFamily(const MonoFunctionPtr& basisFunction)
    : MultiArityFunctionFamily(basisFunction) {}

MonoFunctionPtr
    LeftAssocFunctionFamily::createInstanceFor(const std::vector<SortPtr>& argumentSorts) {
    if (argumentSorts.size() == 1) {
        // TODO: we should not accept arbitrary sorts at this point, even
        // if only a single argument is given
        return new LeftAssocMonoFunction(IndexedString("multiArityOp"),
                                         argumentSorts.front(),
                                         argumentSorts.begin(), argumentSorts.end(),
                                         0);
    }

    SortPtr resSort = argumentSorts.front();
    IndexedString name = "";
    for (int i = 1; i < argumentSorts.size(); ++i) {
        const MonoFunctionPtr basisFun =
            basisFunction->lookup(to_vector(resSort, argumentSorts[i]));
        resSort = basisFun->resultSort;
        name = basisFun->name;
    }

    return new LeftAssocMonoFunction(name, resSort,
                                     argumentSorts.begin(), argumentSorts.end(),
                                     basisFunction);
}

////////////////////////////////////////////////////////////////////////////////

RightAssocFunctionFamily::RightAssocFunctionFamily(const FunctionFamilyPtr& basisFunction)
    : MultiArityFunctionFamily(basisFunction) {}

RightAssocFunctionFamily::RightAssocFunctionFamily(const MonoFunctionPtr& basisFunction)
    : MultiArityFunctionFamily(basisFunction) {}

MonoFunctionPtr
    RightAssocFunctionFamily::createInstanceFor(const std::vector<SortPtr>& argumentSorts) {
    if (argumentSorts.size() == 1) {
        // TODO: we should not accept arbitrary sorts at this point, even
        // if only a single argument is given
        return new RightAssocMonoFunction(IndexedString("multiArityOp"),
                                          argumentSorts.front(),
                                          argumentSorts.begin(), argumentSorts.end(),
                                          0);
    }

    SortPtr resSort = argumentSorts.back();
    IndexedString name = "";
    for (int i = argumentSorts.size() - 2; i >= 0; --i) {
        const MonoFunctionPtr basisFun =
            basisFunction->lookup(to_vector(argumentSorts[i], resSort));
        resSort = basisFun->resultSort;
        name = basisFun->name;
    }

    return new RightAssocMonoFunction(name, resSort,
                                      argumentSorts.begin(), argumentSorts.end(),
                                      basisFunction);
}

////////////////////////////////////////////////////////////////////////////////

ChainableFunctionFamily::ChainableFunctionFamily(const FunctionFamilyPtr& basisFunction,
                                                 const MonoFunctionPtr& fun_true,
                                                 const MonoFunctionPtr& fun_and)
    : MultiArityFunctionFamily(basisFunction),
      fun_true(fun_true), fun_and(fun_and) {}

ChainableFunctionFamily::ChainableFunctionFamily(const MonoFunctionPtr& basisFunction,
                                                 const MonoFunctionPtr& fun_true,
                                                 const MonoFunctionPtr& fun_and)
    : MultiArityFunctionFamily(basisFunction),
      fun_true(fun_true), fun_and(fun_and) {}

MonoFunctionPtr
    ChainableFunctionFamily::createInstanceFor(const std::vector<SortPtr>& argumentSorts) {
    if (!allEqual(argumentSorts.begin(), argumentSorts.end()))
        throw NoSuchSymbolException("");

    // this will throw an exception if the given sorts are not supported by this
    // operator
    const MonoFunctionPtr basisFun =
        basisFunction->lookup(to_vector(argumentSorts.front(),
                                        argumentSorts.front()));

    return new ChainableMonoFunction(basisFun->name, basisFun->resultSort,
                                     argumentSorts.begin(), argumentSorts.end(),
                                     basisFunction, fun_true, fun_and);
}

}
