/**
 * This file is part of the SMT-LIB 2 preprocessor.
 * <http://www.cprover.org/SMT-LIB-LSM/>
 *
 * Copyright (C) 2009 Philipp Ruemmer <ph_r@gmx.net>
 *
 * The SMT-LIB 2 preprocessor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The SMT-LIB 2 preprocessor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the SMT-LIB 2 preprocessor.  If not, see <http://www.gnu.org/licenses/>.
 */


/**
 * Basic functionality for memory management via reference counting
 */

#ifndef REF_COUNTING
#define REF_COUNTING 1

#include <cassert>
#include <functional>
#include <vector>

/**
 * Base class of all reference-counted classes
 */
class RefCounted {

private:
	int m_iRefCount;

public:
	RefCounted();
	virtual ~RefCounted();

	void add_ref();
	void release();
};

/**
 * Smart pointers to reference-counted objects that take care of incrementing
 * and decrementing the counter
 */
template <class T>
class CountingPtr {

private:
	T* m_pData;

public:
    CountingPtr() : m_pData(0) {}

    CountingPtr(T* p) : m_pData(p) {
	    if (m_pData != 0)
	        m_pData->add_ref();
	}

	CountingPtr(const CountingPtr<T>& p) : m_pData((T*)p) {
        if (m_pData != 0)
            m_pData->add_ref();
	}

	~CountingPtr() {
        if (m_pData != 0)
            m_pData->release();
	}

    operator T*() const {
        return m_pData;
    }

    template <class S>
    operator CountingPtr<S>() const {
        return CountingPtr<S>(dynamic_cast<S*>(m_pData));
    }

    template <class S>
    bool isInstanceOf() const {
        return dynamic_cast<S*>(m_pData) != 0;
    }

    template <class S>
    S* asInstanceOf() const {
        return dynamic_cast<S*>(m_pData);
    }

    bool operator== (const CountingPtr<T>& p) const {
        return m_pData == p.m_pData;
    }

    bool operator< (const CountingPtr<T>& p) const {
        return m_pData < p.m_pData;
    }

    bool operator!= (const CountingPtr<T>& p) const {
        return m_pData != p.m_pData;
    }

    CountingPtr<T>& operator=(T* that) {
        if (m_pData != 0)
            m_pData->release();
		m_pData = that;
        if (m_pData != 0)
            m_pData->add_ref();
    	return (*this);
    }

    CountingPtr<T>& operator=(CountingPtr<T>& that) {
        if (m_pData != 0)
            m_pData->release();
		m_pData = that.m_pData;
        if (m_pData != 0)
            m_pData->add_ref();
    	return (*this);
    }

    operator bool() const {
        return m_pData != 0;
    }

    bool operator!() const {
        return m_pData == 0;
    }

    T* operator->() const {
        return m_pData;
    }
};

typedef CountingPtr<RefCounted> RefCountedPtr;

/**
 * Equality predicate with built-in dereferencing
 */
template <class T>
class equal_deref :
	public std::binary_function<CountingPtr<T>, CountingPtr<T>, bool> {

public:
	bool operator() (CountingPtr<T> a, CountingPtr<T> b) {
		return *a == *b;
	}
};

/**
 * Compare two vectors for equality, dereferencing the elements of the vectors
 */
template <class T>
bool equalVectors(const std::vector<CountingPtr<T> >& a,
                  const std::vector<CountingPtr<T> >& b) {
	return a.size() == b.size() &&
		   equal(a.begin(), a.end(), b.begin(), equal_deref<T>());
}

/**
 * Membership predicate with dereferencing
 */
template <class T>
bool contains(const std::vector<CountingPtr<T> >& els,
              const CountingPtr<T>& el) {
    for (typename std::vector<CountingPtr<T> >::const_iterator it = els.begin();
         it != els.end();
         ++it)
        if (**it == *el)
            return true;
    return false;
}

/**
 * Test whether all values in a given range are equal
 */
template <class I> bool allEqual(I begin, I end) {
    if (begin == end)
        return true;

    typename std::iterator_traits<I>::value_type first = *begin;

    ++begin;
    while (begin != end) {
        if (!(*first == **begin))
            return false;
        ++begin;
    }

    return true;
}

#endif
