/**
 * This file is part of the SMT-LIB 2 preprocessor.
 * <http://www.cprover.org/SMT-LIB-LSM/>
 *
 * Copyright (C) 2009 Philipp Ruemmer <ph_r@gmx.net>
 *
 * The SMT-LIB 2 preprocessor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The SMT-LIB 2 preprocessor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the SMT-LIB 2 preprocessor.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "booleanReflector.H"
#include "../smtlib-absyn-visitors.H"
#include "../theories/empty.H"
#include "../theories/function-visitor.H"
#include "../util.H"

namespace SMTBackend {

using namespace std;
using namespace SMTLibAbsyn;
using namespace Visitors;
using namespace Theories;

/**
 * Visitor for replacing Bool with ReflectedBool in the relevant places of an
 * SMT-LIB expression. The boolean argument tells whether an expression is in
 * a term context (occurs as argument of a function), so that reflection is
 * necessary
 */

typedef CountingPtr<SinglyLinkedList<SortPtr> > SortListPtr;

enum ContextKind { OnlyTerms, OnlyFormulae, TermsAndFormulae };

struct ReflectionVisitorArgument {
    ContextKind context;
    SortListPtr boundVariableSorts;

    ReflectionVisitorArgument()
        : context(TermsAndFormulae), boundVariableSorts(0) {}

    ReflectionVisitorArgument(ContextKind context,
                              const SortListPtr& boundVariableSorts)
        : context(context), boundVariableSorts(boundVariableSorts) {}

    ReflectionVisitorArgument setContext(ContextKind c) {
        return ReflectionVisitorArgument(c, boundVariableSorts);
    }
};

/**
 * Determine whether the arguments of some theory function (in SMT-LIB 1)
 * have to be terms or formulae
 */
class FunctionContextComputer : public FunctionVisitor {
public:
    ReflectionVisitorArgument preArgument;
    PreVisitResultPtr res;

    void visit_other        (const FunctionAppPtr& node) {
        res = uniSubArgs(preArgument.setContext(OnlyTerms));
    }

    void visit_fun_not      (const FunctionAppPtr& node) {
        res = uniSubArgs(preArgument.setContext(OnlyFormulae));
    }
    void visit_fun_implies  (const FunctionAppPtr& node) {
        res = uniSubArgs(preArgument.setContext(OnlyFormulae));
    }
    void visit_fun_and      (const FunctionAppPtr& node) {
        res = uniSubArgs(preArgument.setContext(OnlyFormulae));
    }
    void visit_fun_or       (const FunctionAppPtr& node) {
        res = uniSubArgs(preArgument.setContext(OnlyFormulae));
    }
    void visit_fun_xor      (const FunctionAppPtr& node) {
        res = uniSubArgs(preArgument.setContext(OnlyFormulae));
    }
    void visit_polyfun_equal(const FunctionAppPtr& node) {
        if (*node->arguments.front()->sort() == *Sort::Bool)
            res = uniSubArgs(preArgument.setContext(OnlyFormulae));
        else
            res = uniSubArgs(preArgument.setContext(OnlyTerms));
    }
    void visit_polyfun_ite  (const FunctionAppPtr& node) {
        ContextKind branchKind;
        if (*node->arguments.back()->sort() == *Sort::Bool)
            branchKind = OnlyFormulae;
        else
            branchKind = OnlyTerms;

        vector<ReflectionVisitorArgument> arguments;
        arguments.push_back(preArgument.setContext(OnlyFormulae));
        arguments.push_back(preArgument.setContext(branchKind));
        arguments.push_back(preArgument.setContext(branchKind));

        res = subArgs(arguments);
    }
};

class ReflectionPreVisitor
    : public DefaultExpressionVisitor<ReflectionVisitorArgument, int,
                                      PreVisitResultPtr> {

private:
    BooleanReflector& reflector;
    FunctionContextComputer contextComputer;

public:
    ReflectionPreVisitor(BooleanReflector& reflector)
        : reflector(reflector), contextComputer() {}

    void defaultAction(const ExpressionPtr& node) {
        res = uniSubArgs(arg0.setContext(TermsAndFormulae));
    }

    void visitFunctionApp         (const CountingPtr<FunctionApp>& node) {
        contextComputer.preArgument = arg0;
        reflector.benchmark->theory->visit(node, contextComputer);
        res = contextComputer.res;
    }

    void visitLetBinder           (const CountingPtr<LetBinder>& node) {
        vector<ReflectionVisitorArgument> arguments
            (node->variableNames.size(), arg0.setContext(TermsAndFormulae));

        SortListPtr bodyBoundVars = arg0.boundVariableSorts;
        for (unsigned int i = 0; i < node->variableNames.size(); ++i) {
            const SortPtr oriSort = node->variableSorts[i];
            bodyBoundVars =
                new SinglyLinkedList<SortPtr> (
                    reflector.reflectSortArguments(oriSort, false), bodyBoundVars);
        }

        arguments.push_back(ReflectionVisitorArgument(TermsAndFormulae, bodyBoundVars));

        res = subArgs(arguments);
    }

    void visitQuantifiedExpression(const CountingPtr<QuantifiedExpression>& node) {
        SortListPtr bodyBoundVars = arg0.boundVariableSorts;
        for (unsigned int i = 0; i < node->variableNames.size(); ++i) {
            const SortPtr oriSort = node->variableSorts[i];
            const SortPtr newSort = reflector.reflectSortArguments(oriSort, true);
            bodyBoundVars = new SinglyLinkedList<SortPtr> (newSort, bodyBoundVars);
        }

        res = uniSubArgs(ReflectionVisitorArgument(OnlyFormulae, bodyBoundVars));
    }
};

class ReflectionPostVisitor
    : public DefaultExpressionVisitor<ReflectionVisitorArgument,
                                      const vector<ExpressionPtr>*,
                                      ExpressionPtr> {
private:
    BooleanReflector& reflector;

    ExpressionPtr castIfNecessary(const ExpressionPtr& expr) {
        switch (arg0.context) {
        case OnlyTerms:
            if (*expr->sort() == *Sort::Bool)
                return reflector.castToReflectedBool(expr);
            break;
        case OnlyFormulae:
            if (*expr->sort() == *reflector.reflectedBool)
                return reflector.castToBool(expr);
            assert(*expr->sort() == *Sort::Bool);
            break;
        case TermsAndFormulae:
            // nothing
            break;
        }

        return expr;
    }

public:
    ReflectionPostVisitor(BooleanReflector& reflector) : reflector(reflector) {}

    void defaultAction(const ExpressionPtr& node) {
        const ExpressionPtr newExpr = node->update(*arg1);
        res = castIfNecessary(newExpr);
    }

    void visitVariable            (const CountingPtr<Variable>& node) {
        const SortPtr newVarSort =
            arg0.boundVariableSorts->drop(node->index)->value;
        if (*node->sort() == *newVarSort)
            res = node;
        else
            res = new Variable (newVarSort, node->index);
        res = castIfNecessary(res);
    }

    void visitFunctionApp         (const CountingPtr<FunctionApp>& node) {
        const ExpressionPtr newExpr = reflector.updateFunctionApp(node, *arg1);
        res = castIfNecessary(newExpr);
    }

    void visitQuantifiedExpression(const CountingPtr<QuantifiedExpression>& node) {
        bool changed = false;
        vector<SortPtr> newSorts;
        for (unsigned int i = 0; i < node->variableNames.size(); ++i) {
            const SortPtr oriSort = node->variableSorts[i];
            const SortPtr newSort = reflector.reflectSortArguments(oriSort, true);
            if (newSort != oriSort)
                changed = true;
            newSorts.push_back(newSort);
        }

        if (changed || arg1->front() != node->body)
            res = new QuantifiedExpression(node->quantifier,
                                           node->variableNames.begin(),
                                           node->variableNames.end(),
                                           newSorts.begin(), newSorts.end(),
                                           arg1->front());
        else
            res = node;

        res = castIfNecessary(res);
    }
};



////////////////////////////////////////////////////////////////////////////////

BooleanReflector::BooleanReflector(const BenchmarkPtr& benchmark)
    : reflectedBoolCtor(new SortCtor ("ReflectedBool", 0)),
      reflectedBool(reflectedBoolCtor->m()),
      benchmark(benchmark),
      reflectedTrue(new MonoFunction ("reflectedTrue", reflectedBool)),
      reflectedFalse(new MonoFunction ("reflectedFalse", reflectedBool)),
      empty(dynamic_cast<TheoryHelpers*>(benchmark->theory)->getEmpty()) {}

void BooleanReflector::reflect() {
    convertExtraFuns();
    addReflectionSymbols();

    ReflectionPreVisitor preVisitor (*this);
    ReflectionPostVisitor postVisitor (*this);

    for (vector<ExpressionPtr>::iterator it = benchmark->assumptions.begin();
         it != benchmark->assumptions.end();
         ++it)
        *it = collectingVisit(*it, ReflectionVisitorArgument(OnlyFormulae, 0),
                              preVisitor, postVisitor);
    benchmark->formula = collectingVisit(benchmark->formula,
                                         ReflectionVisitorArgument(OnlyFormulae, 0),
                                         preVisitor, postVisitor);
}

void BooleanReflector::convertExtraFuns() {
    for (vector<MonoFunctionPtr>::iterator it = benchmark->extraFuns.begin();
         it != benchmark->extraFuns.end();
         ++it) {
        const MonoFunctionPtr oriFun = *it;

        bool changed = false;
        vector<SortPtr> newArgumentSorts;
        for (vector<SortPtr>::const_iterator it2 = oriFun->argumentSorts.begin();
            it2 != oriFun->argumentSorts.end();
            ++it2) {
            const SortPtr newArg = reflectSortArguments(*it2, true);
            if (newArg != *it2)
                changed = true;
            newArgumentSorts.push_back(newArg);
        }

        const SortPtr newRes = reflectSortArguments(oriFun->resultSort, false);
        if (newRes != oriFun->resultSort)
            changed = true;

        if (changed) {
            const MonoFunctionPtr newFun =
                new MonoFunction (oriFun->name, newRes,
                                  newArgumentSorts.begin(), newArgumentSorts.end());
            extraFunsSubst[oriFun] = newFun;
            *it = newFun;
        }
    }
}

void BooleanReflector::addReflectionSymbols() {
    benchmark->extraSorts.push_back(reflectedBool);
    benchmark->extraFuns.push_back(reflectedTrue);
    benchmark->extraFuns.push_back(reflectedFalse);

    // the reflected truth values are distinct ...
    benchmark->assumptions.push_back(
        empty.fun_not->m(empty.polyfun_equal->m(reflectedTrue->m(),
                                                reflectedFalse->m())));

    // ... and there are only two of them
    benchmark->assumptions.push_back(
        forall("b", reflectedBool,
               empty.fun_or->m(empty.polyfun_equal->m(v(reflectedBool, 0),
                                                      reflectedTrue->m()),
                               empty.polyfun_equal->m(v(reflectedBool, 0),
                                                      reflectedFalse->m()))));
}

ExpressionPtr
    BooleanReflector::updateFunctionApp(const FunctionAppPtr& originalExpression,
                                        const vector<ExpressionPtr>& newArguments) {

    const MonoFunctionPtr oriFun = originalExpression->function;

    // check whether this is an :extrafun that has been replaced with a new one
    const map<MonoFunctionPtr, MonoFunctionPtr>::iterator pos =
        extraFunsSubst.find(oriFun);
    if (pos != extraFunsSubst.end())
        return pos->second->m(newArguments.begin(), newArguments.end());

    MonoFunctionPtr newFun = oriFun;

    // first, check whether the argument sorts have changed and a new function
    // instance needs to be used
    const vector<SortPtr> newArgumentSorts = extractSorts(newArguments);
    if (!equalVectors(newArgumentSorts, oriFun->argumentSorts)) {
        // then we have replace the (polymorphic) function with an instance
        // suitable for the new argument sorts
        const ParentMonoFunctionPtr parentFun = (ParentMonoFunctionPtr)oriFun;
        assert(parentFun);

        try {
            newFun = parentFun->lookupNewInstance(newArgumentSorts);
        } catch (NoSuchSymbolException&) {
            assert(false);
        }
    }

    // now, check whether the result sorts contains parts that need to be
    // reflected (this happens for nullary polymorphic symbols like nil or
    // emptySet)
    const SortPtr newResultSort = reflectSortArguments(newFun->resultSort, false);
    if (!(*newResultSort == *(newFun->resultSort))) {
        const ParentMonoFunctionPtr parentFun = (ParentMonoFunctionPtr)newFun;
        assert(parentFun);

        try {
            newFun = parentFun->lookupNewInstance(newResultSort, newArgumentSorts);
        } catch (NoSuchSymbolException&) {
            assert(false);
        }
    }

    if (newFun == oriFun)
        return originalExpression->update(newArguments);

    return newFun->m(newArguments.begin(), newArguments.end());
}

ExpressionPtr BooleanReflector::castToReflectedBool(const ExpressionPtr& expr) {
    assert(*expr->sort() == *Sort::Bool);
    return empty.polyfun_ite->m(expr, reflectedTrue->m(), reflectedFalse->m());
}

ExpressionPtr BooleanReflector::castToBool(const ExpressionPtr& expr) {
    assert(*expr->sort() == *reflectedBool);
    return empty.polyfun_equal->m(expr, reflectedTrue->m());
}

SortPtr BooleanReflector::reflectSortArguments(const SortPtr& sort,
                                               bool alsoTopLevelOp) {
    if (alsoTopLevelOp && *sort == *Sort::Bool)
        return reflectedBool;

    vector<SortPtr> newArguments;
    bool changed = false;
    for (vector<SortPtr>::const_iterator it = sort->arguments.begin();
         it != sort->arguments.end();
         ++it) {
        const SortPtr newArg = reflectSortArguments(*it, true);
        if (newArg != *it)
            changed = true;
        newArguments.push_back(newArg);
    }

    if (changed)
        return new Sort (sort->op, newArguments.begin(), newArguments.end());
    return sort;
}

}
