/**
 * This file is part of the SMT-LIB 2 preprocessor.
 * <http://www.cprover.org/SMT-LIB-LSM/>
 *
 * Copyright (C) 2009 Philipp Ruemmer <ph_r@gmx.net>
 *
 * The SMT-LIB 2 preprocessor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The SMT-LIB 2 preprocessor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the SMT-LIB 2 preprocessor.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * Eliminate the sort Bool from the benchmarks in places where it is not
 * allowed by SMT-LIB 1: as arguments of functions and as the type of
 * quantified variables. To be on the safe side, also all occurrences of Bool
 * as argument of sort constructors are eliminated (because such occurrences
 * might again lead to illegal function argument sorts or quantified variables
 * when instantiating the theory axioms).
 */

#ifndef BOOLEAN_REFLECTOR
#define BOOLEAN_REFLECTOR 1

#include "../smtlib-absyn.H"
#include "../smtlib-resolver.H"
#include "../theories/poly-function-family.H"
#include "../theories/empty.H"

#include <map>

namespace SMTBackend {

using namespace SMTLibAbsyn;
using namespace Resolver;
using namespace Theories;

class BooleanReflector {
private:
    const SortCtorPtr reflectedBoolCtor;

    /**
     * When doing the conversion, :extrafuns might be replaced with new
     * functions (with the same name, but possibly different argument types)
     */
    std::map<MonoFunctionPtr, MonoFunctionPtr> extraFunsSubst;

    /**
     * Go through the :extrafuns of the benchmark at hand and adjust the
     * argument and result sorts. This also sets up the mapping extraFunsSubst.
     */
    void convertExtraFuns();

    /**
     * Declare the symbols and axioms used for reflecting booleans
     */
    void addReflectionSymbols();

public:
    const SortPtr reflectedBool;

    const BenchmarkPtr benchmark;

private:
    const MonoFunctionPtr reflectedTrue;
    const MonoFunctionPtr reflectedFalse;

    const Empty& empty;

public:
    BooleanReflector(const BenchmarkPtr& benchmark);

    void reflect();

    /**
     * Check whether a MonoFunction has to be replaced with a different function
     * because the argument or result type has changed
     */
    ExpressionPtr updateFunctionApp(const FunctionAppPtr& originalExpression,
                                    const std::vector<ExpressionPtr>& newArguments);

    ExpressionPtr castToReflectedBool(const ExpressionPtr& expr);
    ExpressionPtr castToBool(const ExpressionPtr& expr);

    /**
     * Eliminate occurrences of Bool as arguments of sort constructors
     */
    SortPtr reflectSortArguments(const SortPtr& sort, bool alsoTopLevelOp);
};

}

#endif
