/**
 * This file is part of the SMT-LIB 2 preprocessor.
 * <http://www.cprover.org/SMT-LIB-LSM/>
 *
 * Copyright (C) 2009 Philipp Ruemmer <ph_r@gmx.net>
 *
 * The SMT-LIB 2 preprocessor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The SMT-LIB 2 preprocessor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the SMT-LIB 2 preprocessor.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "translator.H"
#include "booleanReflector.H"
#include "../smtlib-absyn-visitors.H"
#include "../theories/sets.H"
#include "../theories/lists.H"
#include "../theories/maps.H"
#include "../theories/lrm.H"
#include "../theories/function-visitor.H"
#include "../smtlib-printer.H"
#include "../util.H"

#include <sstream>

namespace SMTBackend {

using namespace Visitors;
using namespace Theories;
using namespace std;


////////////////////////////////////////////////////////////////////////////////

class SortCollector : public DefaultExpressionVisitor<int, int, PreVisitResultPtr> {

private:
    Translator& translator;
    const bool declareOtherSorts;

public:
    SortCollector(Translator& translator, bool declareOtherSorts)
		: translator(translator), declareOtherSorts(declareOtherSorts) {}

    void defaultAction(const ExpressionPtr& node) {
        translator.declareSorts(node->sort(), declareOtherSorts);
        res = keepArg();
    }
};

////////////////////////////////////////////////////////////////////////////////

struct PrinterContext {

private:
    unsigned int variableCounter;

public:
    PrinterContext(Translator& translator, ostream& os)
        : variableCounter(0), translator(translator), os(os) {}

    Translator& translator;
    ostream& os;

    string newVariableName() {
        stringstream varName;
        varName << "preprocVar" << variableCounter++;
        return varName.str();
    }
};

/**
 * Visitor to linearise the names of theory function symbols
 *
 * TODO: add special treatment of comparisons like (= emptySet xxx)
 */
class FunctionLineariser : public FunctionVisitor {
private:
    PrinterContext& ctxt;

    void printWithSuffix(const string& prefix, const SortPtr& sort) {
        ctxt.os << prefix;
        addSortSuffix(sort);
    }

    void addSortSuffix(const SortPtr& sort) {
        ctxt.os << "_";
        ctxt.translator.idLineariser.linearise(ctxt.os, sort);
    }

public:
    FunctionLineariser(PrinterContext& ctxt) : ctxt(ctxt) {}

    /**
     * Method called for functions that do not belong to any theory
     */
    void visit_other        (const FunctionAppPtr& node) {
        ctxt.translator.idLineariser.linearise(ctxt.os, node->function->name);
    }
    void visit_polyfun_equal  (const FunctionAppPtr& node) {
    	if (*node->arguments.front()->sort() == *Sort::Bool)
    		ctxt.os << "iff";
    	else
    		ctxt.os << "=";
    }
    void visit_polyfun_ite  (const FunctionAppPtr& node) {
    	if (*node->sort() == *Sort::Bool)
    		ctxt.os << "if_then_else";
    	else
    		ctxt.os << "ite";
    }

    /**
     * The theory of integer arithmetic
     */
    void visit_fun_unary_minus(const FunctionAppPtr& node) {
        ctxt.os << "~";
    }
    void visit_fun_minus      (const FunctionAppPtr& node) {
        ctxt.os << "-";
    }
    void visit_fun_plus       (const FunctionAppPtr& node) {
        ctxt.os << "+";
    }
    void visit_fun_times      (const FunctionAppPtr& node) {
        ctxt.os << "*";
    }
    void visit_fun_leq        (const FunctionAppPtr& node) {
        ctxt.os << "<=";
    }
    void visit_fun_less       (const FunctionAppPtr& node) {
        ctxt.os << "<";
    }
    void visit_fun_geq        (const FunctionAppPtr& node) {
        ctxt.os << ">=";
    }
    void visit_fun_greater    (const FunctionAppPtr& node) {
        ctxt.os << ">";
    }

    /**
     * The theory of finite sets with cardinality
     */
    void visit_polyfun_emptySet(const FunctionAppPtr& node) {
        printWithSuffix("emptySet", node->sort());
    }
    void visit_polyfun_insert  (const FunctionAppPtr& node) {
        printWithSuffix("insert", node->sort());
    }
    void visit_polyfun_in      (const FunctionAppPtr& node) {
        printWithSuffix("in", node->arguments.back()->sort());
    }
    void visit_polyfun_subset  (const FunctionAppPtr& node) {
        printWithSuffix("subset", node->arguments.back()->sort());
    }
    void visit_polyfun_union   (const FunctionAppPtr& node) {
        printWithSuffix("union", node->sort());
    }
    void visit_polyfun_inter   (const FunctionAppPtr& node) {
        printWithSuffix("inter", node->sort());
    }
    void visit_polyfun_setminus(const FunctionAppPtr& node) {
        printWithSuffix("setminus", node->sort());
    }
    void visit_polyfun_card    (const FunctionAppPtr& node) {
        printWithSuffix("card", node->arguments.back()->sort());
    }

    /**
     * The theory of lists with length
     */
    void visit_polyfun_nil   (const FunctionAppPtr& node) {
        printWithSuffix("nil", node->sort());
    }
    void visit_polyfun_cons  (const FunctionAppPtr& node) {
        printWithSuffix("cons", node->sort());
    }
    void visit_polyfun_head  (const FunctionAppPtr& node) {
        printWithSuffix("head", node->arguments.back()->sort());
    }
    void visit_polyfun_tail  (const FunctionAppPtr& node) {
        printWithSuffix("tail", node->arguments.back()->sort());
    }
    void visit_polyfun_append(const FunctionAppPtr& node) {
        printWithSuffix("append", node->sort());
    }
    void visit_polyfun_length(const FunctionAppPtr& node) {
        printWithSuffix("length", node->arguments.back()->sort());
    }
    void visit_polyfun_nth   (const FunctionAppPtr& node) {
        printWithSuffix("nth", node->arguments.back()->sort());
    }

    /**
     * The theory of finite maps
     */
    void visit_polyfun_emptyMap (const FunctionAppPtr& node) {
        printWithSuffix("emptyMap", node->sort());
    }
    void visit_polyfun_apply    (const FunctionAppPtr& node) {
        printWithSuffix("apply", node->arguments.front()->sort());
    }
    void visit_polyfun_overwrite(const FunctionAppPtr& node) {
        printWithSuffix("overwrite", node->sort());
    }
    void visit_polyfun_domain   (const FunctionAppPtr& node) {
        printWithSuffix("domain", node->arguments.front()->sort());
    }
    void visit_polyfun_range    (const FunctionAppPtr& node) {
        printWithSuffix("range", node->arguments.front()->sort());
    }
    void visit_polyfun_restrict (const FunctionAppPtr& node) {
        printWithSuffix("restrict", node->sort());
    }
    void visit_polyfun_subtract (const FunctionAppPtr& node) {
        printWithSuffix("subtract", node->sort());
    }

    /**
     * The theory of finite relations with cardinality
     */
    void visit_polyfun_tuple    (const FunctionAppPtr& node) {
        ctxt.translator.idLineariser.linearise(ctxt.os, node->function->name);
        ctxt.os << "_";
        ctxt.translator.idLineariser.linearise(ctxt.os, node->sort());
    }
    void visit_polyfun_project  (const FunctionAppPtr& node) {
        ctxt.translator.idLineariser.linearise(ctxt.os, node->function->name);
        ctxt.os << "_";
        ctxt.translator.idLineariser.linearise(ctxt.os, node->arguments.front()->sort());
    }
    void visit_polyfun_product  (const FunctionAppPtr& node) {
        ctxt.translator.idLineariser.linearise(ctxt.os, node->function->name);
        ctxt.os << "_";
        ctxt.translator.idLineariser.linearise(ctxt.os, node->sort());
    }

    /**
     * The theory of lists, relations (including sets), and partial maps
     */
    void visit_polyfun_inds     (const FunctionAppPtr& node) {
        printWithSuffix("inds", node->arguments.front()->sort());
    }
    void visit_polyfun_elems    (const FunctionAppPtr& node) {
        printWithSuffix("elems", node->arguments.front()->sort());
    }
};

/**
 * Visitors for linearising SMT-LIB expressions. The string argument denotes
 * a separator that to be added to the output stream immediately before an
 * expression.
 */
class PrinterPreVisitor
    : public ArgExpressionVisitor<PrinterArgument, int, PreVisitResultPtr> {
private:
    PrinterContext& ctxt;
    FunctionLineariser functionLineariser;

public:
    PrinterPreVisitor(PrinterContext& ctxt)
         : ctxt(ctxt), functionLineariser(ctxt) {}

    void visitVariable            (const CountingPtr<Variable>& node) {
        ctxt.os << arg0.prefix;
        ctxt.os << arg0.boundVariables->drop(node->index)->value;
        res = keepArg();
    }

    void visitLiteral             (const CountingPtr<Literal>& node) {
        ctxt.os << arg0.prefix;
        ctxt.os << node->value;
        res = keepArg();
    }

    void visitFunctionApp         (const CountingPtr<FunctionApp>& node) {
        ctxt.os << arg0.prefix;
        if (node->arity() > 0)
            ctxt.os << "(";
        ctxt.translator.theory->visit(node, functionLineariser);
        // function arguments are separated by spaces
        res = uniSubArgs(PrinterArgument(arg0.boundVariables, " "));
    }

    void visitLetBinder           (const CountingPtr<LetBinder>& node) {
    	assert(*node->sort() == *Sort::Bool); // TODO

        ctxt.os << arg0.prefix;

        vector<PrinterArgument> arguments;
        StringListPtr bodyBoundVars = arg0.boundVariables;

        for (unsigned int i = 0; i < node->variableNames.size(); ++i) {
            const bool flet = *node->variableSorts[i] == *Sort::Bool;
            const string newVarName = (flet ? "$" : "?") + ctxt.newVariableName();
            const string letPrefix = flet ? "(flet (" : "(let (";
            const string completePrefix =
                (i > 0 ? ") " : "") + letPrefix + newVarName + " ";

            arguments.push_back(PrinterArgument(arg0.boundVariables, completePrefix));

            bodyBoundVars = new SinglyLinkedList<string> (newVarName, bodyBoundVars);
        }

        arguments.push_back(PrinterArgument(bodyBoundVars, ") "));

        res = subArgs(arguments);
    }

    void visitQuantifiedExpression(const CountingPtr<QuantifiedExpression>& node) {
        ctxt.os << arg0.prefix;

        switch (node->quantifier) {
        case Forall:
            ctxt.os << "(forall ";
            break;
        case Exists:
            ctxt.os << "(exists ";
            break;
        }

        StringListPtr bodyBoundVars = arg0.boundVariables;

        for (unsigned int i = 0; i < node->variableNames.size(); ++i) {
            assert(!(*node->variableSorts[i] == *Sort::Bool));

            const string varName = "?" + node->variableNames[i];

            ctxt.os << "(" << varName << " ";
            ctxt.translator.idLineariser.linearise(ctxt.os, node->variableSorts[i]);
            ctxt.os << ") ";

            bodyBoundVars = new SinglyLinkedList<string> (varName, bodyBoundVars);
        }

        res = uniSubArgs(PrinterArgument(bodyBoundVars, ""));
    }
};

class PrinterPostVisitor
    : public ArgExpressionVisitor<PrinterArgument, const std::vector<int>*, int> {
private:
    PrinterContext& ctxt;

public:
    PrinterPostVisitor(PrinterContext& ctxt) : ctxt(ctxt) {}

    void visitVariable            (const CountingPtr<Variable>& node) {}

    void visitLiteral             (const CountingPtr<Literal>& node) {}

    void visitFunctionApp         (const CountingPtr<FunctionApp>& node) {
        if (node->arity() > 0)
            ctxt.os << ")";
    }

    void visitLetBinder           (const CountingPtr<LetBinder>& node) {
        for (unsigned int i = 0; i < node->variableNames.size(); ++i)
            ctxt.os << ")";
    }

    void visitQuantifiedExpression(const CountingPtr<QuantifiedExpression>& node) {
    	ctxt.os << ")";
    }
};

////////////////////////////////////////////////////////////////////////////////

Translator::Translator(const FileTemplates& fileTemplates,
                       std::ostream& stream,
                       const Resolver::BenchmarkPtr& benchmark)
    : fileTemplates(fileTemplates), os(stream), benchmark(benchmark),
      relationAxiomGenerator(stream), theory(benchmark->theory) {
	setupSorts();
}

void Translator::linearise(const ExpressionPtr& expr) {
    PrinterContext ctxt (*this, os);
    PrinterPreVisitor preVisitor (ctxt);
    PrinterPostVisitor postVisitor (ctxt);

    collectingVisit(expr, PrinterArgument(0, ""), preVisitor, postVisitor);
}

void Translator::translate() {
    BooleanReflector booleanReflector (benchmark);
    booleanReflector.reflect();

	os << "(benchmark " << benchmark->name << endl;
	os << ":logic AUFLIA" << endl;
// TODO: add better information about the pre-processor
	os << ":note { File generated by the SMT-LIB 2 pre-preprocessor }"
	   << endl << endl;

	declareOtherSorts();
	os << endl;

    instantiateAxioms();
    os << endl;

	declareExtraFuns();
	os << endl;

    for (vector<ExpressionPtr>::iterator it = benchmark->assumptions.begin();
         it != benchmark->assumptions.end();
         ++it) {
        os << ":assumption" << endl;
        linearise(*it);
        os << endl;
    }
    os << endl;

    os << ":formula" << endl;
    linearise(benchmark->formula);
    os << endl << ")" << endl;
}

void Translator::declareOtherSorts() {
    SortCollector collector (*this, true);
    forallExpressions(collector);
    declareExtraFunSorts(true);
}

void Translator::instantiateAxioms() {
    SortCollector collector (*this, false);
    forallExpressions(collector);
    declareExtraFunSorts(false);
}

void Translator::declareExtraFunSorts(bool declareOtherSorts) {
    for (vector<MonoFunctionPtr>::iterator it = benchmark->extraFuns.begin();
         it != benchmark->extraFuns.end();
         ++it) {
    	const MonoFunctionPtr fun = *it;

    	declareSorts(fun->resultSort, declareOtherSorts);
    	for (vector<SortPtr>::const_iterator it = fun->argumentSorts.begin();
    	     it != fun->argumentSorts.end();
    	     ++it)
    		declareSorts(*it, declareOtherSorts);
    }
}

void Translator::declareSorts(const SortPtr& sort, bool declareOtherSorts) {
    // First declare possible sort arguments (the order in which sorts are
    // declared in SMT-LIB 1 matters)
    for (vector<SortPtr>::const_iterator it = sort->arguments.begin();
         it != sort->arguments.end();
         ++it)
        declareSorts(*it, declareOtherSorts);

	if (sort->op == sort_Set) {
    	if (!declareOtherSorts)
    		instantiateSetAxioms(sort);
	} else if (sort->op == sort_List) {
	    if (!declareOtherSorts)
	        instantiateListAxioms(sort);
    } else if (sort->op == sort_Map) {
        if (!declareOtherSorts)
            instantiateMapAxioms(sort);
    } else if (sort_Tuple && sort_Tuple->isTupleSortCtor(sort->op)) {
        if (!declareOtherSorts)
            instantiateRelationAxioms(sort);
	} else if (sort->op == sort_Bool) {
		// nothing
	} else if (sort->op == sort_Int) {
		// nothing
	} else {
        if (declareOtherSorts)
        	declareOtherSort(sort);
	}
}

void Translator::declareExtraFuns() {
    for (vector<MonoFunctionPtr>::iterator it = benchmark->extraFuns.begin();
         it != benchmark->extraFuns.end();
         ++it) {
    	const MonoFunctionPtr fun = *it;
    	const bool predicate = (*fun->resultSort == *Sort::Bool);

    	if (predicate)
    		os << ":extrapreds ((";
    	else
    		os << ":extrafuns ((";

    	idLineariser.linearise(os, fun->name);

    	for (vector<SortPtr>::const_iterator it = fun->argumentSorts.begin();
    	     it != fun->argumentSorts.end();
    	     ++it) {
    	    // boolean arguments should have been eliminated at an earlier stage
    		assert(!(**it == *Sort::Bool));
    		os << " ";
    		idLineariser.linearise(os, *it);
    	}

    	if (!predicate) {
    		os << " ";
    		idLineariser.linearise(os, fun->resultSort);
    	}

    	os << "))" << endl;
    }
}

void Translator::setupSorts() {
    if (Sets* sets = dynamic_cast<Sets*>(theory)) {
        sort_Bool = sets->empty.sort_Bool;
        sort_Int = sets->integers.sort_Int;
        sort_Set = sets->sort_Set;
    } else if (Lists* lists = dynamic_cast<Lists*>(theory)) {
        sort_Bool = lists->empty.sort_Bool;
        sort_Int = lists->integers.sort_Int;
        sort_List = lists->sort_List;
    } else if (Maps* maps = dynamic_cast<Maps*>(theory)) {
        sort_Bool = maps->empty.sort_Bool;
        sort_Int = maps->sets.integers.sort_Int;
        sort_Set = maps->sets.sort_Set;
        sort_Map = maps->sort_Map;
    } else if (Relations* relations = dynamic_cast<Relations*>(theory)) {
        sort_Bool = relations->empty.sort_Bool;
        sort_Int = relations->sets.integers.sort_Int;
        sort_Set = relations->sets.sort_Set;
        sort_Tuple = relations->sort_Tuple;
    } else if (LRM* lrm = dynamic_cast<LRM*>(theory)) {
        sort_Bool = lrm->empty.sort_Bool;
        sort_Int = lrm->integers.sort_Int;
        sort_Set = lrm->sets.sort_Set;
        sort_Tuple = lrm->relations.sort_Tuple;
        sort_Map = lrm->maps.sort_Map;
        sort_List = lrm->lists.sort_List;
    } else {
        assert(false);
    }
}

void Translator::declareOtherSort(const SortPtr& sort) {
	if (!contains(declaredOtherSorts, sort)) {
		os << ":extrasorts (";
		idLineariser.linearise(os, sort);
		os << ")" << endl;

		declaredOtherSorts.push_back(sort);
	}
}

void Translator::instantiateTEmbedding(const SortPtr& sort) {
    if (contains(instantiatedTEmbeddings, sort))
        return;

    StringSubstitution subst;
    subst["<T>"] = idLineariser.toString(sort);

    fileTemplates.T_embedding.instantiate(os, subst);
    instantiatedTEmbeddings.push_back(sort);
}

void Translator::instantiateSetAxioms(const SortPtr& sort) {
    if (instantiatedSetSorts.empty())
        fileTemplates.Set_template.instantiate(os, StringSubstitution());

    if (!contains(instantiatedSetSorts, sort)) {
        instantiatedSetSorts.push_back(sort);

        const SortPtr argSort = sort->arguments.front();
        assert(!(*argSort == *Sort::Bool));

        instantiateTEmbedding(argSort);

        StringSubstitution subst;
        subst["<T>"] = idLineariser.toString(argSort);
        subst["<Set_T>"] = idLineariser.toString(sort);

        fileTemplates.Set_T_template.instantiate(os, subst);

        if (sort_Tuple && sort_Tuple->isTupleSortCtor(argSort->op)) {
            for (vector<SortPtr>::const_iterator it = argSort->arguments.begin();
                 it != argSort->arguments.end();
                 ++it)
                instantiateSetAxioms(sort_Set->m(*it));
            relationAxiomGenerator.instantiateSetTupleAxioms(sort);
        }
    }
}

void Translator::instantiateListAxioms(const SortPtr& sort) {
    if (!contains(instantiatedListSorts, sort)) {
        const SortPtr argSort = sort->arguments.front();
        assert(!(*argSort == *Sort::Bool));

        StringSubstitution subst;
        subst["<T>"] = idLineariser.toString(argSort);
        subst["<List_T>"] = idLineariser.toString(sort);

        fileTemplates.List_T_template.instantiate(os, subst);

        if (sort_Set) {
        	// also introduce the functions inds and elems

        	instantiateSetAxioms(sort_Set->m(sort_Int->m()));
        	instantiateSetAxioms(sort_Set->m(argSort));

            subst["<Set_T>"] = idLineariser.toString(sort_Set->m(argSort));
        	fileTemplates.List_Set_T_template.instantiate(os, subst);
        }

        instantiatedListSorts.push_back(sort);
    }
}

void Translator::instantiateMapAxioms(const SortPtr& sort) {
    if (!contains(instantiatedMapSorts, sort)) {
        const SortPtr arg0Sort = sort->arguments.front();
        const SortPtr arg1Sort = sort->arguments.back();
        assert(!(*arg0Sort == *Sort::Bool) && !(*arg1Sort == *Sort::Bool));

        instantiateSetAxioms(sort_Set->m(arg0Sort));
        instantiateSetAxioms(sort_Set->m(arg1Sort));

        StringSubstitution subst;
        subst["<S>"] = idLineariser.toString(arg0Sort);
        subst["<T>"] = idLineariser.toString(arg1Sort);
        subst["<Set_S>"] = idLineariser.toString(sort_Set->m(arg0Sort));
        subst["<Set_T>"] = idLineariser.toString(sort_Set->m(arg1Sort));
        subst["<Map_S_T>"] = idLineariser.toString(sort);

        fileTemplates.Map_S_T_template.instantiate(os, subst);
        instantiatedMapSorts.push_back(sort);
    }
}

void Translator::instantiateRelationAxioms(const SortPtr& sort) {
    if (!contains(instantiatedRelationSorts, sort)) {
        relationAxiomGenerator.instantiateTupleAxioms(sort);
        instantiatedRelationSorts.push_back(sort);
    }
}

}
