/**
 * This file is part of the SMT-LIB 2 preprocessor.
 * <http://www.cprover.org/SMT-LIB-LSM/>
 *
 * Copyright (C) 2009 Philipp Ruemmer <ph_r@gmx.net>
 *
 * The SMT-LIB 2 preprocessor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The SMT-LIB 2 preprocessor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the SMT-LIB 2 preprocessor.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * Translate benchmarks into the SMT-LIB 1 format, using axioms and
 * uninterpreted functions to encode the theories
 */

#ifndef SMT_TRANSLATOR
#define SMT_TRANSLATOR 1

#include "file-templates.H"
#include "id-lineariser.H"
#include "relation-axioms.H"
#include "../smtlib-resolver.H"
#include "../smtlib-absyn-visitors.H"
#include "../theories/theory.H"
#include "../theories/relations.H"

#include <iostream>
#include <vector>
#include <string>
#include <map>
#include <set>

namespace SMTBackend {

using namespace SMTLibAbsyn;
using namespace Visitors;

class Translator {

private:
    const FileTemplates& fileTemplates;
    std::ostream& os;
    const Resolver::BenchmarkPtr benchmark;
    RelationAxiomGenerator relationAxiomGenerator;

    /**
     * List of uninterpreted sorts that have already been declared in the
     * SMT-LIB output
     */
    std::vector<SortPtr> declaredOtherSorts;

    /**
     * The instantiation of theory axioms is driven by the sorts that are found
     * in the benchmark. The following lists contain those sorts that have
     * already been handled
     */
    std::vector<SortPtr> instantiatedSetSorts;
    std::vector<SortPtr> instantiatedListSorts;
    std::vector<SortPtr> instantiatedMapSorts;
    std::vector<SortPtr> instantiatedRelationSorts;
    std::vector<SortPtr> instantiatedTEmbeddings;

    /**
     * Apply some visitor to all :assumptions and the :formula in the benchmark
     */
    template <class A>
    void forallExpressions(ArgExpressionVisitor<A, int, PreVisitResultPtr >& visitor) {
        PassiveExpressionVisitor<int, const std::vector<int>*, int> postVisitor;

    	for (std::vector<ExpressionPtr>::iterator it = benchmark->assumptions.begin();
             it != benchmark->assumptions.end();
             ++it)
            collectingVisit(*it, 0, visitor, postVisitor);
        collectingVisit(benchmark->formula, 0, visitor, postVisitor);
    }

    /**
     * Assign the fields sort_Bool, sort_Int, etc.
     */
    void setupSorts();

    /**
     * Instantiate the theory axioms for all sorts that occur in the benchmark
     */
    void instantiateAxioms();

    /**
     * Declare all uninterpreted sorts occurring in the benchmark
     */
    void declareOtherSorts();

    void declareExtraFunSorts(bool declareOtherSorts);
    void declareOtherSort(const SortPtr& sort);
    void instantiateSetAxioms(const SortPtr& sort);
    void instantiateListAxioms(const SortPtr& sort);
    void instantiateMapAxioms(const SortPtr& sort);
    void instantiateRelationAxioms(const SortPtr& sort);
    void instantiateTEmbedding(const SortPtr& sort);

    /**
     * Declare all uninterpreted functions occurring in the benchmark
     */
    void declareExtraFuns();

    /**
     * Linearise an SMT-LIB expression, assuming that it has been transformed
     * to a form that corresponds to SMT-LIB 1
     */
    void linearise(const ExpressionPtr& expr);

public:
    IdLineariser idLineariser;

    Theories::Theory* const theory;

    SortCtorPtr sort_Bool;
    SortCtorPtr sort_Int;
    SortCtorPtr sort_Set;
    SortCtorPtr sort_List;
    SortCtorPtr sort_Map;
    Theories::TupleSortCtorPtr sort_Tuple;

    Translator(const FileTemplates& fileTemplates,
               std::ostream& stream,
               const Resolver::BenchmarkPtr& benchmark);

    void translate();

    /**
     * Called by the visitor for collecting sorts occurring in expressions
     */
    void declareSorts(const SortPtr& sort, bool declareOtherSorts);

};

}

#endif
