/**
 * This file is part of the SMT-LIB 2 preprocessor.
 * <http://www.cprover.org/SMT-LIB-LSM/>
 *
 * Copyright (C) 2009 Philipp Ruemmer <ph_r@gmx.net>
 *
 * The SMT-LIB 2 preprocessor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The SMT-LIB 2 preprocessor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the SMT-LIB 2 preprocessor.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * Various visitor classes that can be used to analyse, traverse, and modify
 * SMT-LIB expressions
 */

#ifndef SMTLIB_ABSYN_VISITORS
#define SMTLIB_ABSYN_VISITORS 1

#include <vector>
#include <stack>

#include "smtlib-absyn.H"

namespace Visitors {
using namespace SMTLibAbsyn;

/**
 * Because virtual functions cannot be polymorphic, there is apparently no
 * elegant way to realise visitors with arbitrary parameter and result types.
 * Instead, in the following abstract visitor class parameters and results are
 * passed through fields.
 */
template <class A0, class A1, class R>
class ArgExpressionVisitor : public ExpressionVisitor {
public:
	/**
	 * The two arguments of the visitor, to be set before calling one of the
	 * visit-methods
	 */
	A0 arg0; A1 arg1;

	/**
	 * The result of calling one of the visit-methods, assigned in the
	 * visit-methods
	 */
	R res;

	R operator()(const ExpressionPtr& node) {
		node->visit(*this);
		return res;
	}

	R operator()(const ExpressionPtr& node, A0 argument0) {
		arg0 = argument0;
		node->visit(*this);
		return res;
	}

	R operator()(const ExpressionPtr& node, A0 argument0, A1 argument1) {
		arg0 = argument0;
		arg1 = argument1;
		node->visit(*this);
		return res;
	}
};

template <class A0, class A1, class R>
class DefaultExpressionVisitor : public ArgExpressionVisitor<A0, A1, R> {
    virtual void defaultAction(const ExpressionPtr& node) = 0;

    void visitVariable            (const CountingPtr<Variable>& node) {
        defaultAction(node);
    }
    void visitLiteral             (const CountingPtr<Literal>& node) {
        defaultAction(node);
    }
    void visitFunctionApp         (const CountingPtr<FunctionApp>& node) {
        defaultAction(node);
    }
    void visitLetBinder           (const CountingPtr<LetBinder>& node) {
        defaultAction(node);
    }
    void visitQuantifiedExpression(const CountingPtr<QuantifiedExpression>& node) {
        defaultAction(node);
    }
};

/**
 * Visitor that does nothing
 */
template <class A0, class A1, class R>
class PassiveExpressionVisitor : public DefaultExpressionVisitor<A0, A1, R> {
    void defaultAction(const ExpressionPtr& node) {
        // nothing
    }
};

////////////////////////////////////////////////////////////////////////////////
// Visitors for traversing an expression

enum PreVisitResultKind {
	KeepArg, TryAgain, UniSubArgs, SubArgs, ShortCut
};

class PreVisitResult : public RefCounted {
public:
	virtual PreVisitResultKind kind() = 0;
};

/**
 * Use the same argument for the direct sub-expressions as for this expression
 */
class KeepArgResult : public PreVisitResult {
public:
	PreVisitResultKind kind();
	static const CountingPtr<KeepArgResult> INSTANCE;
};

CountingPtr<KeepArgResult> keepArg();

/**
 * Call preVisitor again with a different expression and argument
 */
template <class A>
class TryAgainResult : public PreVisitResult {
public:
	PreVisitResultKind kind() {
		return TryAgain;
	}

	TryAgainResult(const ExpressionPtr& newNode, A newArg)
		: newNode(newNode), newArg(newArg) {}

	const ExpressionPtr newNode;
	const A newArg;
};

template <class A>
CountingPtr<TryAgainResult<A> > tryAgain(const ExpressionPtr& newNode, A newArg) {
	return new TryAgainResult<A>(newNode, newArg);
}

/**
 * Use arg for each of the direct sub-expressions
 */
template <class A>
class UniSubArgsResult : public PreVisitResult {
public:
	PreVisitResultKind kind() {
		return UniSubArgs;
	}

	UniSubArgsResult(A arg) : arg(arg) {}

	const A arg;
};

template <class A>
CountingPtr<UniSubArgsResult<A> > uniSubArgs(A arg) {
	return new UniSubArgsResult<A>(arg);
}

/**
 * Specify the arguments to use for the individual sub-expressions
 */
template <class A>
class SubArgsResult : public PreVisitResult {
public:
	PreVisitResultKind kind() {
		return SubArgs;
	}

	SubArgsResult(const std::vector<A>& args) : args(args) {}

	const std::vector<A> args;
};

template <class A>
CountingPtr<SubArgsResult<A> > subArgs(const std::vector<A>& args) {
	return new SubArgsResult<A>(args);
}

/**
 * Skip the call to postVisit and do not visit any of the
 * sub-expressions. Instead, directly return res as result
 */
template <class R>
class ShortCutResult : public PreVisitResult {
public:
	PreVisitResultKind kind() {
		return ShortCut;
	}

	ShortCutResult(R res) : res(res) {}

	const R res;
};

template <class R>
CountingPtr<ShortCutResult<R> > shortCut(R res) {
	return new ShortCutResult<R>(res);
}

typedef CountingPtr<PreVisitResult> PreVisitResultPtr;

/**
 * Internal subclass of the expression AST, to remember points on the stack
 * where it is necessary to call the postVisitor
 */
template <class A>
class PostVisitExpression : public Expression {
public:
    PostVisitExpression(const ExpressionPtr& expression, A arg)
        : expression(expression), arg(arg) {}

    SortPtr sort() const {
        return expression->sort();
    }

    const ExpressionPtr expression;
    const A arg;
};

/**
 * Visitor schema that traverses an expression in depth-first left-first order.
 * For each node, the visitor preVisitor is called when descending
 * and the visitor postVisitor when returning. The visitor works
 * with iteration (not recursion) and is able to deal also with large
 * expressions. The second argument of the preVisitor is not used
 */
template <class A, class R>
R collectingVisit(const ExpressionPtr& expression,
                  A arg,
                  ArgExpressionVisitor<A, int, PreVisitResultPtr >& preVisitor,
                  ArgExpressionVisitor<A, const std::vector<R>*, R>& postVisitor) {
    std::stack<ExpressionPtr> toVisit;
    std::stack<A> argsToVisit;
    std::stack<R> results;

    toVisit.push(expression);
    argsToVisit.push(arg);

    std::vector<R> subRes;

    while (!toVisit.empty()) {
        const ExpressionPtr expr = toVisit.top();
        toVisit.pop();
        const CountingPtr<PostVisitExpression<A> > postVisitExpr = expr;

        if (postVisitExpr) {

            for (int i = postVisitExpr->expression->arity() - 1; i >= 0; --i) {
                subRes.insert(subRes.begin(), results.top());
                results.pop();
            }

            results.push(postVisitor(postVisitExpr->expression,
                                     postVisitExpr->arg, &subRes));

            subRes.clear();

        } else {

            const A arg = argsToVisit.top();
            argsToVisit.pop();

            const PreVisitResultPtr preVisitRes = preVisitor(expr, arg);
            const PreVisitResultKind kind = preVisitRes->kind();

            switch (kind) {

            case ShortCut:
                // directly push the result, skip the call to postVisit and the
                // recursive calls
                results.push(preVisitRes.asInstanceOf<ShortCutResult<R> >()->res);
                break;

            case TryAgain:
                toVisit.push(preVisitRes.asInstanceOf<TryAgainResult<A> >()->newNode);
                argsToVisit.push(preVisitRes.asInstanceOf<TryAgainResult<A> >()->newArg);
                break;

            default:
                if (expr->arity() > 0) {
                    // recurse
                    toVisit.push(new PostVisitExpression<A>(expr, arg));
                    for (int i = expr->arity() - 1; i >= 0; --i)
                        toVisit.push(expr->subExpr(i));

                    switch (kind) {

                    case KeepArg:
                        for (int i = expr->arity() - 1; i >= 0; --i)
                            argsToVisit.push(arg);
                        break;

                    case UniSubArgs:
                        {
                            const A newArg =
                                preVisitRes.asInstanceOf<UniSubArgsResult<A> >()->arg;
                            for (int i = expr->arity() - 1; i >= 0; --i)
                                argsToVisit.push(newArg);
                        }
                        break;

                    case SubArgs:
                        {
                            const std::vector<A>& args =
                                preVisitRes.asInstanceOf<SubArgsResult<A> >()->args;
                            assert(args.size() == expr->arity());
                            for (int i = expr->arity() - 1; i >= 0; --i)
                                argsToVisit.push(args[i]);
                        }
                        break;

                    default:
                        assert(false); // not reachable
                        break;
                    }
                } else {
                    // otherwise, we can directly call the postVisitor
                    results.push(postVisitor(expr, arg, &subRes));
                }
                break;
            }

        }
    }

    assert(results.size() == 1);
    return results.top();
}

}

#endif
