/**
 * This file is part of the SMT-LIB 2 preprocessor.
 * <http://www.cprover.org/SMT-LIB-LSM/>
 *
 * Copyright (C) 2009 Philipp Ruemmer <ph_r@gmx.net>
 *
 * The SMT-LIB 2 preprocessor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The SMT-LIB 2 preprocessor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the SMT-LIB 2 preprocessor.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <vector>

#include "sort-unification.H"
#include "util.H"

namespace SortUnification {

using namespace std;
using namespace SMTLibAbsyn;

SortVariable::SortVariable()
    : SortCtor(IndexedString(""), 0) {}

////////////////////////////////////////////////////////////////////////////////

set<SortVariablePtr> domain(const Substitution& subst) {

    set<SortVariablePtr> res;
    for (Substitution::const_iterator it = subst.begin();
         it != subst.end();
         ++it) {
        if (it->second->op != it->first)
            res.insert(it->first);
    }

    return res;
}

set<SortVariablePtr> codomain(const Substitution& subst) {

    set<SortVariablePtr> res;
    for (Substitution::const_iterator it = subst.begin();
         it != subst.end();
         ++it) {
        if (it->second->op != it->first)
            insertAll(res, freeVariables(it->second));
    }

    return res;
}

bool idempotent(const Substitution& subst) {
    const set<SortVariablePtr> dom = domain(subst);
    const set<SortVariablePtr> codom = codomain(subst);

    return disjoint(dom, codom);
}

////////////////////////////////////////////////////////////////////////////////

bool addAssignment(const SortVariablePtr& var, const SortPtr& sort,
                   Substitution& unifier);

bool unify(const SortPtr& a, const SortPtr& b, Substitution& unifier) {
    assert(idempotent(unifier));

    if (a == b)
        return true;

    const SortPtr aSubst = (SortVariablePtr)a->op ? substitute(a, unifier) : a;
    const SortPtr bSubst = (SortVariablePtr)b->op ? substitute(b, unifier) : b;

    if (const SortVariablePtr aVar = (SortVariablePtr)aSubst->op) {
        return addAssignment(aVar, bSubst, unifier);
    } else {
        if (const SortVariablePtr bVar = (SortVariablePtr)bSubst->op) {
            return addAssignment(bVar, aSubst, unifier);
        } else {
            if (aSubst->op != bSubst->op)
                return false;

            for (unsigned int i = 0; i < aSubst->arguments.size(); ++i) {
                if (!unify(aSubst->arguments[i], bSubst->arguments[i], unifier))
                    return false;
            }

            assert(idempotent(unifier));
            return true;
        }
    }
}

bool addAssignment(const SortVariablePtr& var, const SortPtr& sort,
                   Substitution& unifier) {
    assert(unifier.find(var) == unifier.end());

    if (sort->op == var)
        return true;

    // occurs check
    const set<SortVariablePtr> freeVars = freeVariables(sort);

    if (freeVars.find(var) != freeVars.end())
        return false;

    // make the unifier idempotent
    Substitution smallSubst;
    smallSubst[var] = sort;

    for (Substitution::iterator it = unifier.begin(); it != unifier.end(); ++it)
        unifier[it->first] = substitute(it->second, smallSubst);

    unifier[var] = sort;

    assert(idempotent(unifier));
    return true;
}

// TODO: more efficient implementation, if necessary
set<SortVariablePtr> freeVariables(const SortPtr& sort) {
    set<SortVariablePtr> res;

    if (const SortVariablePtr sortVar = (SortVariablePtr)sort->op) {
        res.insert(sortVar);
        return res;
    }

    // otherwise, recurse
    for (vector<SortPtr>::const_iterator it = sort->arguments.begin();
         it != sort->arguments.end();
         ++it)
        insertAll(res, freeVariables(*it));

    return res;
}

SortPtr substitute(const SortPtr& sort, const Substitution& subst) {
    if (const SortVariablePtr sortVar = (SortVariablePtr)sort->op) {
        const Substitution::const_iterator pos = subst.find(sortVar);
        if (pos == subst.end())
            return sort;
        else
            return pos->second;
    }

    // otherwise, recurse
    vector<SortPtr> newArguments;
    for (vector<SortPtr>::const_iterator it = sort->arguments.begin();
         it != sort->arguments.end();
         ++it)
        newArguments.push_back(substitute(*it, subst));

    return sort->update(newArguments);
}

}
