/**
 * This file is part of the SMT-LIB 2 preprocessor.
 * <http://www.cprover.org/SMT-LIB-LSM/>
 *
 * Copyright (C) 2009 Philipp Ruemmer <ph_r@gmx.net>
 *
 * The SMT-LIB 2 preprocessor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The SMT-LIB 2 preprocessor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the SMT-LIB 2 preprocessor.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <vector>
#include <set>

#include "empty.H"
#include "../util.H"
#include "poly-function-family.H"
#include "function-visitor.H"
#include "../multi-arity-ops.H"

namespace Theories {

using namespace SMTLibAbsyn;
using namespace std;


void TheoryHelpers::registerMonoFunction(Environment &env,
                                         const MonoFunctionPtr& function) const {
    env.addFunctionFamily(function->name.body,
                          new MonoFunctionFamily(function), true);
}

void TheoryHelpers::registerPolyFunction(Environment &env,
                                         const PolyFunctionFamilyPtr& function) const {
    env.addFunctionFamily(function->name.body, function, true);
}

void TheoryHelpers::registerMultiArityFunction
        (Environment &env,
         const MonoFunctionPtr& basisFunction,
         const MultiArityFunctionFamilyPtr& function) const {
    env.addFunctionFamily(basisFunction->name.body, function, true);
}

void TheoryHelpers::registerMultiArityFunction
        (Environment &env,
         const PolyFunctionFamilyPtr& basisFunction,
         const MultiArityFunctionFamilyPtr& function) const {
    env.addFunctionFamily(basisFunction->name.body, function, true);
}

////////////////////////////////////////////////////////////////////////////////

MonoFunctionPtr DistinctFunction::lookup(const std::vector<std::string>& indexes,
                                         const std::vector<SortPtr>& argumentSorts) {
    if (indexes.size() != 0 || !allEqual(argumentSorts.begin(), argumentSorts.end()))
        throw NoSuchSymbolException("");

    // check whether the requested instance already exists
    for (vector<MonoFunctionPtr>::iterator it = instances.begin();
         it != instances.end();
         ++it) {
        if (equalVectors((*it)->argumentSorts, argumentSorts))
            return *it;
    }

    const MonoFunctionPtr res =
        new ParentMonoFunction (IndexedString("distinct", vector<string>()),
                                Sort::Bool,
                                argumentSorts.begin(), argumentSorts.end(),
                                this);
    instances.push_back(res);

    return res;
}

MonoFunctionPtr DistinctFunction::lookup(const std::vector<std::string>& indexes,
                                         const SortPtr& resSort,
                                         const std::vector<SortPtr>& argumentSorts) {
    const MonoFunctionPtr res = lookup(indexes, argumentSorts);

    if (!(*res->resultSort == *resSort))
        throw NoSuchSymbolException("");

    return res;
}

////////////////////////////////////////////////////////////////////////////////

PolyFunctionFamilyPtr createEqual() {
    const SortVariablePtr alpha = new SortVariable;

    vector<SortPtr> arguments;
    arguments.push_back(alpha->m());
    arguments.push_back(alpha->m());

    return new PolyFunctionFamily("=", Sort::Bool, arguments);
}

PolyFunctionFamilyPtr createITE() {
    const SortVariablePtr alpha = new SortVariable;

    vector<SortPtr> arguments;
    arguments.push_back(Sort::Bool);
    arguments.push_back(alpha->m());
    arguments.push_back(alpha->m());

    return new PolyFunctionFamily("ite", alpha->m(), arguments);
}

Empty::Empty()
    : sort_Bool    (SortCtor::Bool),
      fun_true     (new MonoFunction ("true", Sort::Bool)),
      fun_false    (new MonoFunction ("false", Sort::Bool)),
      fun_not      (new MonoFunction ("not", Sort::Bool, to_vector(Sort::Bool))),
      fun_implies  (new MonoFunction ("implies", Sort::Bool,
                                      to_vector(Sort::Bool, Sort::Bool))),
      fun_and      (new MonoFunction ("and", Sort::Bool,
                                      to_vector(Sort::Bool, Sort::Bool))),
      fun_or       (new MonoFunction ("or", Sort::Bool,
                                      to_vector(Sort::Bool, Sort::Bool))),
      fun_xor      (new MonoFunction ("xor", Sort::Bool,
                                      to_vector(Sort::Bool, Sort::Bool))),
      polyfun_equal(createEqual()),
      polyfun_ite  (createITE()),
      polyfun_distinct(new DistinctFunction),
      multiArityFun_implies (new RightAssocFunctionFamily (fun_implies)),
      multiArityFun_and     (new LeftAssocFunctionFamily (fun_and)),
      multiArityFun_or      (new LeftAssocFunctionFamily (fun_or)),
      multiArityFun_xor     (new LeftAssocFunctionFamily (fun_xor)),
      multiArityFun_equal   (new ChainableFunctionFamily ((FunctionFamily*)polyfun_equal,
                                                          fun_true, fun_and))
      {}

void Empty::registerSymbolsLocally(Environment& env) const {
    env.addSortCtorFamily(sort_Bool->name.body,
                          new SimpleSortCtorFamily(sort_Bool), true);
    registerMonoFunction(env, fun_true);
    registerMonoFunction(env, fun_false);
    registerMonoFunction(env, fun_not);
    registerPolyFunction(env, polyfun_ite);
    env.addFunctionFamily("distinct", polyfun_distinct, true);
    registerMultiArityFunction(env, fun_implies, multiArityFun_implies);
    registerMultiArityFunction(env, fun_and, multiArityFun_and);
    registerMultiArityFunction(env, fun_or, multiArityFun_or);
    registerMultiArityFunction(env, fun_xor, multiArityFun_xor);
    registerMultiArityFunction(env, polyfun_equal, multiArityFun_equal);
}

void Empty::registerSymbols(Environment& env) const {
    registerSymbolsLocally(env);
}

void Empty::visit(const FunctionAppPtr& node,
                  FunctionVisitor& visitor) const {
    if (!visitLocally(node, visitor))
        visitor.visit_other(node);
}

bool Empty::visitLocally(const FunctionAppPtr& node,
                         FunctionVisitor& visitor) const {
    const MonoFunctionPtr fun = node->function;
    if (fun == fun_true)
        visitor.visit_fun_true(node);
    else if (fun == fun_false)
        visitor.visit_fun_false(node);
    else if (fun == fun_not)
        visitor.visit_fun_not(node);
    else if (fun == fun_implies)
        visitor.visit_fun_implies(node);
    else if (fun == fun_and)
        visitor.visit_fun_and(node);
    else if (fun == fun_or)
        visitor.visit_fun_or(node);
    else if (fun == fun_xor)
        visitor.visit_fun_xor(node);
    else if (fun == fun_false)
        visitor.visit_fun_false(node);
    else {
        const ParentMonoFunctionPtr parentFun = (ParentMonoFunctionPtr)fun;
        if (!parentFun)
            return false;
        if (parentFun->parent == polyfun_equal)
            visitor.visit_polyfun_equal(node);
        else if (parentFun->parent == polyfun_ite)
            visitor.visit_polyfun_ite(node);
        else if (parentFun->parent == polyfun_distinct)
            visitor.visit_polyfun_distinct(node);
        else
            return false;
    }
    return true;
}

const Empty& Empty::getEmpty() const {
    return *this;
}

}
