/**
 * This file is part of the SMT-LIB 2 preprocessor.
 * <http://www.cprover.org/SMT-LIB-LSM/>
 *
 * Copyright (C) 2009 Philipp Ruemmer <ph_r@gmx.net>
 *
 * The SMT-LIB 2 preprocessor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The SMT-LIB 2 preprocessor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the SMT-LIB 2 preprocessor.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "lrm.H"
#include "../util.H"
#include "function-visitor.H"

namespace Theories {

using namespace SMTLibAbsyn;
using namespace std;

PolyFunctionFamilyPtr createElems(const SortCtorPtr& sort_Set,
                                  const SortCtorPtr& sort_List) {
    const SortVariablePtr alpha = new SortVariable;
    return new PolyFunctionFamily("elems", sort_Set->m(alpha->m()),
                                  to_vector(sort_List->m(alpha->m())));
}

LRM::LRM(const Empty&     empty,
         const Integers&  integers,
         const Sets&      sets,
         const Lists&     lists,
         const Relations& relations,
         const Maps&      maps)
 : empty(empty), integers(integers), sets(sets), lists(lists),
   relations(relations), maps(maps),
   polyfun_inds (new PolyFunctionFamily("inds",
                     sets.sort_Set->m(integers.sort_Int->m()),
                     to_vector(lists.sort_List->m((new SortVariable)->m())))),
   polyfun_elems(createElems(sets.sort_Set, lists.sort_List))
 {}

void LRM::registerSymbolsLocally(Environment& env) const {
    registerPolyFunction(env, polyfun_inds);
    registerPolyFunction(env, polyfun_elems);
}

void LRM::registerSymbols(Environment& env) const {
    empty.registerSymbolsLocally(env);
    integers.registerSymbolsLocally(env);
    sets.registerSymbolsLocally(env);
    lists.registerSymbolsLocally(env);
    relations.registerSymbolsLocally(env);
    maps.registerSymbolsLocally(env);
    registerSymbolsLocally(env);
}

void LRM::visit(const FunctionAppPtr& node,
                FunctionVisitor& visitor) const {
    if (!(visitLocally(node, visitor) ||
          empty.visitLocally(node, visitor) ||
          integers.visitLocally(node, visitor) ||
          sets.visitLocally(node, visitor) ||
          lists.visitLocally(node, visitor) ||
          relations.visitLocally(node, visitor) ||
          maps.visitLocally(node, visitor)))
        visitor.visit_other(node);
}

bool LRM::visitLocally(const FunctionAppPtr& node,
                       FunctionVisitor& visitor) const {
    const ParentMonoFunctionPtr parentFun = (ParentMonoFunctionPtr)node->function;
    if (!parentFun)
        return false;
    const FunctionFamilyPtr parent = parentFun->parent;
    if (parent == polyfun_inds)
        visitor.visit_polyfun_inds(node);
    else if (parent == polyfun_elems)
        visitor.visit_polyfun_elems(node);
    else
        return false;
    return true;
}

LRM* LRM::create() {
    Empty* empty = new Empty;
    Integers* integers = new Integers (*empty);
    Sets* sets = new Sets (*empty, *integers);
    Lists* lists = new Lists (*empty, *integers);
    Relations* relations = new Relations (*empty, *sets);
    Maps* maps = new Maps (*empty, *sets);
    return new LRM (*empty, *integers, *sets, *lists, *relations, *maps);
}

const Empty& LRM::getEmpty() const {
    return empty;
}

}
