/**
 * This file is part of the SMT-LIB 2 preprocessor.
 * <http://www.cprover.org/SMT-LIB-LSM/>
 *
 * Copyright (C) 2009 Philipp Ruemmer <ph_r@gmx.net>
 *
 * The SMT-LIB 2 preprocessor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The SMT-LIB 2 preprocessor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the SMT-LIB 2 preprocessor.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "relations.H"
#include "../util.H"
#include "function-visitor.H"
#include "../smtlib-resolver.H"
#include "poly-function-family.H"

#include <cstdlib>

namespace Theories {

using namespace SMTLibAbsyn;
using namespace Resolver;
using namespace std;

SortCtorPtr TupleSortCtor::lookup(const std::vector<std::string>& indexes) {
    if (indexes.size() != 1)
        throw NoSuchSymbolException("");

    const int arity = atoi(indexes.front().c_str());
    if (arity < 1)
        throw NoSuchSymbolException("");

    map<unsigned int, SortCtorPtr>::iterator pos = instances.find(arity);
    if (pos != instances.end())
        return pos->second;

    const SortCtorPtr res = new SortCtor (IndexedString("Tuple", indexes), arity);
    instances[(unsigned int)arity] = res;

    return res;
}

bool TupleSortCtor::isTupleSortCtor(const SortCtorPtr& op) {
    map<unsigned int, SortCtorPtr>::iterator pos = instances.find(op->arity);
    return pos != instances.end() && pos->second == op;
}

////////////////////////////////////////////////////////////////////////////////

TupleFunctionFamily::TupleFunctionFamily(const TupleSortCtorPtr& sort_Tuple)
    : sort_Tuple(sort_Tuple) {}

MonoFunctionPtr TupleFunctionFamily::lookupInstance
                            (const std::vector<SortPtr>& argumentSorts) {
    // check whether the requested instance already exists
    for (vector<MonoFunctionPtr>::iterator it = instances.begin();
         it != instances.end();
         ++it) {
        if (equalVectors((*it)->argumentSorts, argumentSorts))
            return *it;
    }
    return 0;
}

MonoFunctionPtr TupleFunctionFamily::lookup(const std::vector<std::string>& indexes,
                                            const SortPtr& resSort,
                                            const std::vector<SortPtr>& argumentSorts) {
    const MonoFunctionPtr res = lookup(indexes, argumentSorts);

    if (!(*res->resultSort == *resSort))
        throw NoSuchSymbolException("");

    return res;
}

////////////////////////////////////////////////////////////////////////////////

TupleCtor::TupleCtor(const TupleSortCtorPtr& sort_Tuple)
    : TupleFunctionFamily(sort_Tuple) {}

MonoFunctionPtr TupleCtor::lookup(const std::vector<std::string>& indexes,
                                  const std::vector<SortPtr>& argumentSorts) {
    // this will throw an exception if no tuple constructor with the given
    // indexes exist
    const SortCtorPtr tupleSortCtor = sort_Tuple->lookup(indexes);

    if (tupleSortCtor->arity != argumentSorts.size())
        throw NoSuchSymbolException("");

    MonoFunctionPtr res = lookupInstance(argumentSorts);
    if (res)
        return res;

    const SortPtr resultSort =
        new Sort (tupleSortCtor, argumentSorts.begin(), argumentSorts.end());
    res = new ParentMonoFunction (IndexedString("tuple", indexes), resultSort,
                                  argumentSorts.begin(), argumentSorts.end(),
                                  this);
    instances.push_back(res);

    return res;
}

////////////////////////////////////////////////////////////////////////////////

TupleProjection::TupleProjection(const TupleSortCtorPtr& sort_Tuple)
    : TupleFunctionFamily(sort_Tuple) {}

MonoFunctionPtr TupleProjection::lookup(const std::vector<std::string>& indexes,
                                        const std::vector<SortPtr>& argumentSorts) {
    if (indexes.size() != 2)
        throw NoSuchSymbolException("");

    const int arity = atoi(indexes.front().c_str());
    const int index = atoi(indexes.back().c_str());
    if (arity < 1 || index < 1 || index > arity)
        throw NoSuchSymbolException("");

    if (argumentSorts.size() != 1)
        throw NoSuchSymbolException("");

    const SortCtorPtr tupleOp = argumentSorts.front()->op;

    if (!sort_Tuple->isTupleSortCtor(tupleOp) ||
        tupleOp->arity != (unsigned int)arity)
        throw NoSuchSymbolException("");

    // check whether the requested instance already exists
    for (vector<MonoFunctionPtr>::iterator it = instances.begin();
         it != instances.end();
         ++it) {
        if (equalVectors((*it)->argumentSorts, argumentSorts) &&
            (*it)->name.indexes == indexes)
            return *it;
    }

    const SortPtr resultSort = argumentSorts.front()->arguments[index - 1];
    const MonoFunctionPtr res =
        new ParentMonoFunction (IndexedString("project", indexes), resultSort,
                                argumentSorts.begin(), argumentSorts.end(),
                                this);
    instances.push_back(res);

    return res;
}

////////////////////////////////////////////////////////////////////////////////

TupleProduct::TupleProduct(const TupleSortCtorPtr& sort_Tuple,
                           const SortCtorPtr& sort_Set)
    : TupleFunctionFamily(sort_Tuple), sort_Set(sort_Set) {}

MonoFunctionPtr TupleProduct::lookup(const std::vector<std::string>& indexes,
                                     const std::vector<SortPtr>& argumentSorts) {
    // this will throw an exception if no tuple constructor with the given
    // indexes exist
    const SortCtorPtr tupleSortCtor = sort_Tuple->lookup(indexes);

    const int arity = atoi(indexes.front().c_str());
    if ((unsigned int)arity != argumentSorts.size())
        throw NoSuchSymbolException("");

    MonoFunctionPtr res = lookupInstance(argumentSorts);
    if (res)
        return res;

    vector<SortPtr> argElementSorts;
    for (vector<SortPtr>::const_iterator it = argumentSorts.begin();
         it != argumentSorts.end();
         ++it) {
        if (!(*it)->op == sort_Set)
            throw NoSuchSymbolException("");
        argElementSorts.push_back((*it)->arguments.front());
    }

    const SortPtr resultElementSort =
        new Sort (tupleSortCtor, argElementSorts.begin(), argElementSorts.end());
    const SortPtr resultSort =
        sort_Set->m(resultElementSort);

    res = new ParentMonoFunction (IndexedString("product", indexes), resultSort,
                                  argumentSorts.begin(), argumentSorts.end(),
                                  this);
    instances.push_back(res);

    return res;
}

////////////////////////////////////////////////////////////////////////////////

Relations::Relations(const Empty& empty, const Sets& sets)
: empty(empty), sets(sets),
  sort_Tuple      (new TupleSortCtor),
  polyfun_tuple   (new TupleCtor (sort_Tuple)),
  polyfun_project (new TupleProjection (sort_Tuple)),
  polyfun_product (new TupleProduct (sort_Tuple, sets.sort_Set))
  {}

void Relations::registerSymbolsLocally(Environment& env) const {
    env.addSortCtorFamily("Tuple", sort_Tuple, true);
    env.addFunctionFamily("tuple", polyfun_tuple, true);
    env.addFunctionFamily("project", polyfun_project, true);
    env.addFunctionFamily("product", polyfun_product, true);
}

void Relations::registerSymbols(Environment& env) const {
    empty.registerSymbolsLocally(env);
    sets.registerSymbolsLocally(env);
    sets.integers.registerSymbolsLocally(env);
    registerSymbolsLocally(env);
}

void Relations::visit(const FunctionAppPtr& node,
                      FunctionVisitor& visitor) const {
    if (!(visitLocally(node, visitor) ||
          empty.visitLocally(node, visitor) ||
          sets.integers.visitLocally(node, visitor) ||
          sets.visitLocally(node, visitor)))
        visitor.visit_other(node);
}

bool Relations::visitLocally(const FunctionAppPtr& node,
                             FunctionVisitor& visitor) const {
    const ParentMonoFunctionPtr parentFun = (ParentMonoFunctionPtr)node->function;
    if (!parentFun)
        return false;
    const FunctionFamilyPtr parent = parentFun->parent;
    if (parent == polyfun_tuple)
        visitor.visit_polyfun_tuple(node);
    else if (parent == polyfun_project)
        visitor.visit_polyfun_project(node);
    else if (parent == polyfun_product)
        visitor.visit_polyfun_product(node);
    else
        return false;
    return true;
}

Relations* Relations::create() {
    Empty* empty = new Empty;
    Integers* integers = new Integers (*empty);
    Sets* sets = new Sets (*empty, *integers);
    return new Relations (*empty, *sets);
}

const Empty& Relations::getEmpty() const {
    return empty;
}

}
