/**
 * This file is part of the SMT-LIB 2 preprocessor.
 * <http://www.cprover.org/SMT-LIB-LSM/>
 *
 * Copyright (C) 2009 Philipp Ruemmer <ph_r@gmx.net>
 *
 * The SMT-LIB 2 preprocessor is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The SMT-LIB 2 preprocessor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with the SMT-LIB 2 preprocessor.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "sets.H"
#include "../util.H"
#include "function-visitor.H"

namespace Theories {

using namespace SMTLibAbsyn;
using namespace std;

PolyFunctionFamilyPtr createIn(const SortCtorPtr& sort_Set) {
    const SortVariablePtr alpha = new SortVariable;
    return new PolyFunctionFamily("in", Sort::Bool,
                                  to_vector(alpha->m(), sort_Set->m(alpha->m())));
}

PolyFunctionFamilyPtr createInsert(const SortCtorPtr& sort_Set) {
    const SortVariablePtr alpha = new SortVariable;
    const SortPtr Set_alpha = sort_Set->m(alpha->m());

    return new PolyFunctionFamily("insert", Set_alpha,
                                  to_vector(alpha->m(), Set_alpha));
}

PolyFunctionFamilyPtr createSubset(const SortCtorPtr& sort_Set) {
    const SortVariablePtr alpha = new SortVariable;
    const SortPtr Set_alpha = sort_Set->m(alpha->m());

    return new PolyFunctionFamily("subset", Sort::Bool,
                                  to_vector(Set_alpha, Set_alpha));
}

PolyFunctionFamilyPtr createBinOp(const string& name, const SortCtorPtr& sort_Set) {
    const SortVariablePtr alpha = new SortVariable;
    const SortPtr Set_alpha = sort_Set->m(alpha->m());

    return new PolyFunctionFamily(name, Set_alpha,
                                  to_vector(Set_alpha, Set_alpha));
}

Sets::Sets(const Empty& empty, const Integers& integers)
: empty(empty), integers(integers),
  sort_Set        (new SortCtor ("Set", 1)),
  polyfun_emptySet(new PolyFunctionFamily ("emptySet",
                                           sort_Set->m((new SortVariable)->m()),
                                           vector<SortPtr>())),
  polyfun_insert  (createInsert(sort_Set)),
  polyfun_in      (createIn(sort_Set)),
  polyfun_subset  (createSubset(sort_Set)),
  polyfun_union   (createBinOp("union", sort_Set)),
  polyfun_inter   (createBinOp("inter", sort_Set)),
  polyfun_setminus(createBinOp("setminus", sort_Set)),
  polyfun_card    (new PolyFunctionFamily("card", integers.sort_Int->m(),
                                          to_vector(sort_Set->m((new SortVariable)->m())))),
  multiArityFun_insert  (new RightAssocFunctionFamily ((FunctionFamily*)polyfun_insert)),
  multiArityFun_subset  (new ChainableFunctionFamily  ((FunctionFamily*)polyfun_subset,
                                                       empty.fun_true, empty.fun_and)),
  multiArityFun_union   (new LeftAssocFunctionFamily ((FunctionFamily*)polyfun_union)),
  multiArityFun_inter   (new LeftAssocFunctionFamily ((FunctionFamily*)polyfun_inter)),
  multiArityFun_setminus(new LeftAssocFunctionFamily ((FunctionFamily*)polyfun_setminus))
  {}

void Sets::registerSymbolsLocally(Environment& env) const {
    env.addSortCtorFamily(sort_Set->name.body,
                          new SimpleSortCtorFamily(sort_Set), true);
    registerPolyFunction(env, polyfun_emptySet);
    registerPolyFunction(env, polyfun_in);
    registerPolyFunction(env, polyfun_card);
    registerMultiArityFunction(env, polyfun_insert, multiArityFun_insert);
    registerMultiArityFunction(env, polyfun_subset, multiArityFun_subset);
    registerMultiArityFunction(env, polyfun_union, multiArityFun_union);
    registerMultiArityFunction(env, polyfun_inter, multiArityFun_inter);
    registerMultiArityFunction(env, polyfun_setminus, multiArityFun_setminus);
}

void Sets::registerSymbols(Environment& env) const {
    integers.registerSymbolsLocally(env);
    empty.registerSymbolsLocally(env);
    registerSymbolsLocally(env);
}

void Sets::visit(const FunctionAppPtr& node,
                 FunctionVisitor& visitor) const {
    if (!(visitLocally(node, visitor) ||
          empty.visitLocally(node, visitor) ||
          integers.visitLocally(node, visitor)))
        visitor.visit_other(node);
}

bool Sets::visitLocally(const FunctionAppPtr& node,
                        FunctionVisitor& visitor) const {
    const ParentMonoFunctionPtr parentFun = (ParentMonoFunctionPtr)node->function;
    if (!parentFun)
        return false;
    const FunctionFamilyPtr parent = parentFun->parent;
    if (parent == polyfun_emptySet)
        visitor.visit_polyfun_emptySet(node);
    else if (parent == polyfun_insert)
        visitor.visit_polyfun_insert(node);
    else if (parent == polyfun_in)
        visitor.visit_polyfun_in(node);
    else if (parent == polyfun_subset)
        visitor.visit_polyfun_subset(node);
    else if (parent == polyfun_union)
        visitor.visit_polyfun_union(node);
    else if (parent == polyfun_inter)
        visitor.visit_polyfun_inter(node);
    else if (parent == polyfun_setminus)
        visitor.visit_polyfun_setminus(node);
    else if (parent == polyfun_card)
        visitor.visit_polyfun_card(node);
    else
        return false;
    return true;
}

Sets* Sets::create() {
    Empty* empty = new Empty;
    Integers* integers = new Integers (*empty);
    return new Sets (*empty, *integers);
}

const Empty& Sets::getEmpty() const {
    return empty;
}

}
